use std::path::{Path, PathBuf};

use clap::Args;

use crate::prelude::*;
use crate::arrow::polars::*;
use crate::ids::codes::{NS_GR_WORK, NS_GR_BOOK};

use polars::prelude::*;

#[derive(Args, Debug)]
pub struct CICommand {
  /// Cluster ratings
  #[arg(long="ratings")]
  ratings: bool,

  /// Cluster add-to-shelf actions
  #[arg(long="add-actions")]
  add_actions: bool,

  /// Cluster using simple data instead of full data.
  #[arg(long="simple")]
  simple: bool,

  /// Cluster using native GoodReads works instead of book clusters.
  #[arg(long="native-works")]
  native_works: bool,

  /// Write output to FILE
  #[arg(short='o', long="output", name="FILE")]
  output: PathBuf,
}

#[derive(Debug, PartialEq, Eq)]
enum SrcType {
  Simple,
  Full,
}

#[derive(Debug, PartialEq, Eq)]
enum ActionType {
  Ratings,
  AddActions,
}

#[derive(Debug, PartialEq, Eq)]
enum AggType {
  Clusters,
  NativeWorks,
}

#[derive(Debug)]
pub struct ClusterOp {
  actions: ActionType,
  data: SrcType,
  clusters: AggType,
  output: PathBuf,
}

impl CICommand {
  pub fn exec(&self) -> Result<()> {
    let mut op = if self.add_actions {
      ClusterOp::add_actions(&self.output)
    } else if self.ratings {
      ClusterOp::ratings(&self.output)
    } else {
      error!("must specify one of --add-actions or --raitngs");
      return Err(anyhow!("no operating mode specified"));
    };
    if self.native_works {
      op = op.native_works();
    }
    if self.simple {
      op = op.simple();
    }

    op.cluster()
  }
}

impl ClusterOp {
  /// Start a new action-clustering operation.
  pub fn add_actions<P: AsRef<Path>>(path: P) -> ClusterOp {
    ClusterOp {
      actions: ActionType::AddActions,
      data: SrcType::Full,
      clusters: AggType::Clusters,
      output: path.as_ref().to_path_buf(),
    }
  }

  /// Start a new rating-clustering operation.
  pub fn ratings<P: AsRef<Path>>(path: P) -> ClusterOp {
    ClusterOp {
      actions: ActionType::Ratings,
      data: SrcType::Full,
      clusters: AggType::Clusters,
      output: path.as_ref().to_path_buf(),
    }
  }

  /// Set operation to cluster simple records instead of full records.
  pub fn simple(self) -> ClusterOp {
    ClusterOp {
      data: SrcType::Simple,
      ..self
    }
  }

  /// Set operation to cluster with native works instead of clusters.
  pub fn native_works(self) -> ClusterOp {
    ClusterOp {
      clusters: AggType::NativeWorks,
      ..self
    }
  }

  /// Run the clustering operation.
  pub fn cluster(self) -> Result<()> {
    let interactions = self.load_input()?;
    let interactions = self.filter(interactions);
    let interactions = self.project_and_sort(interactions);
    let actions = interactions.clone().groupby(&[
      col("user"), col("item")
    ]).agg(self.aggregates());

    let actions = self.maybe_integrate_ratings(actions, &interactions);

    debug!("logical plan: {:?}", actions.describe_plan());
    debug!("optimized plan: {:?}", actions.describe_optimized_plan()?);
    info!("collecting results");
    let actions = actions.collect()?;

    info!("writing {} actions to {:?}", actions.height(), &self.output);
    save_df_parquet(actions, &self.output)?;

    Ok(())
  }

  /// Load the input.
  fn load_input(&self) -> PolarsResult<LazyFrame> {
    let dir = match self.data {
      SrcType::Full => "full",
      SrcType::Simple => "simple",
    };
    let path = format!("goodreads/{}/gr-interactions.parquet", dir);
    let data = LazyFrame::scan_parquet(path, Default::default())?;

    let links = LazyFrame::scan_parquet("goodreads/gr-book-link.parquet", Default::default())?;

    let data = data.join(links, &[col("book_id")], &[col("book_id")], JoinType::Inner);
    Ok(data)
  }

  /// Filter the data frame to only the actions we want
  fn filter(&self, frame: LazyFrame) -> LazyFrame {
    match self.actions {
      ActionType::Ratings => {
        frame.filter(col("rating").is_not_null())
      },
      _ => frame
    }
  }

  /// Create an identity column reference.
  fn id_col(&self) -> Expr {
    match self.clusters {
      AggType::Clusters => {
        info!("grouping by integrated clusters");
        col("cluster")
      },
      AggType::NativeWorks => {
        info!("grouping by native works");
        when(
          col("work_id").is_not_null()
        ).then(
          col("work_id") + lit(NS_GR_WORK.base())
        ).otherwise(
          col("book_id") + lit(NS_GR_BOOK.base())
        )
      },
    }
  }

  /// Project and sort (if possible) the data.
  fn project_and_sort(&self, frame: LazyFrame) -> LazyFrame {
    match self.data {
      SrcType::Simple => {
        frame.select(&[
          col("user_id").alias("user"),
          self.id_col().alias("item"),
          col("rating"),
        ])
      },
      SrcType::Full => {
        frame.select(&[
          col("user_id").alias("user"),
          self.id_col().alias("item"),
          (col("updated").cast(DataType::Int64) / lit(1000)).alias("timestamp"),
          col("rating"),
        ])
      }
    }
  }

  /// Aggreate the interactions.
  fn aggregates(&self) -> Vec<Expr> {
    match (&self.actions, &self.data) {
      (ActionType::Ratings, SrcType::Simple) => {
        vec![
          col("rating").median().alias("rating"),
          col("item").count().alias("nratings"),
        ]
      },
      (ActionType::Ratings, SrcType::Full) => {
        vec![
          col("rating").median().alias("rating"),
          col("rating").last().alias("last_rating"),
          col("timestamp").min().alias("first_time"),
          col("timestamp").max().alias("last_time"),
          col("item").count().alias("nratings"),
        ]
      },
      (ActionType::AddActions, SrcType::Simple) => {
        vec![
          col("item").count().alias("nactions"),
        ]
      },
      (ActionType::AddActions, SrcType::Full) => {
        vec![
          col("timestamp").min().alias("first_time"),
          col("timestamp").max().alias("last_time"),
          col("item").count().alias("nactions"),
        ]
      },
    }
  }

  fn maybe_integrate_ratings(&self, actions: LazyFrame, source: &LazyFrame) -> LazyFrame {
    match &self.actions {
      ActionType::AddActions => {
        let ratings = source.clone().filter(col("rating").is_not_null());
        let ratings = ratings.groupby(["user", "item"]).agg(&[
          col("rating").last().alias("last_rating")
        ]);
        actions.join(ratings, &[col("user"), col("item")], &[col("user"), col("item")], JoinType::Left)
      },
      _ => actions
    }
  }
}
